@@ -24,9 +24,14 @@ typedef struct {
 } luv_dir_t;
 #endif
 
+typedef struct {
+  uv_fs_t* req;
+} luv_fs_scandir_t;
+
 static uv_fs_t* luv_check_fs(lua_State* L, int index) {
-  if (luaL_testudata(L, index, "uv_fs") != NULL) {
-    return (uv_fs_t*)lua_touserdata(L, index);
+  if (luaL_testudata(L, index, "uv_fs_scandir") != NULL) {
+    luv_fs_scandir_t* scandir_req = (luv_fs_scandir_t*)lua_touserdata(L, index);
+    return (uv_fs_t*)(scandir_req->req);
   }
   uv_fs_t* req = (uv_fs_t*)luaL_checkudata(L, index, "uv_req");
   luaL_argcheck(L, req->type == UV_FS && req->data, index, "Expected uv_fs_t");
@@ -251,6 +256,13 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
   }
 
   if (req->result < 0) {
+    if (req->fs_type == UV_FS_SCANDIR) {
+      // We need to unref the luv_fs_scandir_t userdata to allow it to be garbage collected.
+      // The scandir callback can only be called once, so we now know that the
+      // req can be safely garbage collected.
+      luaL_unref(L, LUA_REGISTRYINDEX, data->data_ref);
+      data->data_ref = LUA_NOREF;
+    }
     lua_pushnil(L);
     if (fs_req_has_dest_path(req)) {
       lua_rawgeti(L, LUA_REGISTRYINDEX, data->data_ref);
@@ -336,8 +348,15 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
       return 1;
 
     case UV_FS_SCANDIR:
-      // Expose the userdata for the request.
-      lua_rawgeti(L, LUA_REGISTRYINDEX, data->req_ref);
+      // The luv_fs_scandir_t userdata is stored in data_ref.
+      // We want to return this instead of the uv_req_t because the
+      // luv_fs_scandir_t userdata has a gc method.
+      lua_rawgeti(L, LUA_REGISTRYINDEX, data->data_ref);
+      // We now want to unref the userdata to allow it to be garbage collected.
+      // The scandir callback can only be called once, so we now know that the
+      // req can be safely garbage collected.
+      luaL_unref(L, LUA_REGISTRYINDEX, data->data_ref);
+      data->data_ref = LUA_NOREF;
       return 1;
 
 #if LUV_UV_VERSION_GEQ(1, 28, 0)
@@ -363,6 +382,9 @@ static int push_fs_result(lua_State* L, uv_fs_t* req) {
       return 1;
     }
     case UV_FS_READDIR: {
+      luaL_unref(L, LUA_REGISTRYINDEX, data->data_ref);
+      data->data_ref = LUA_NOREF;
+
       if(req->result > 0) {
         size_t i;
         uv_dir_t *dir = (uv_dir_t*)req->ptr;
@@ -428,14 +450,14 @@ static void luv_fs_cb(uv_fs_t* req) {
 // variable 'nargs' to the number of return values
 #define FS_CALL_NORETURN(func, req, ...) {                \
   int ret, sync;                                          \
-  luv_req_t* data = (luv_req_t*)req->data;                \
-  sync = data->callback_ref == LUA_NOREF;                 \
-  ret = uv_fs_##func(data->ctx->loop, req, __VA_ARGS__,   \
+  luv_req_t* lreq = (luv_req_t*)req->data;                \
+  sync = lreq->callback_ref == LUA_NOREF;                 \
+  ret = func(lreq->ctx->loop, req, __VA_ARGS__,           \
                      sync ? NULL : luv_fs_cb);            \
   if (req->fs_type != UV_FS_ACCESS && ret < 0) {          \
     lua_pushnil(L);                                       \
     if (fs_req_has_dest_path(req)) {                      \
-      lua_rawgeti(L, LUA_REGISTRYINDEX, data->data_ref);  \
+      lua_rawgeti(L, LUA_REGISTRYINDEX, lreq->data_ref);  \
       const char* dest_path = lua_tostring(L, -1);        \
       lua_pop(L, 1);                                      \
       lua_pushfstring(L, "%s: %s: %s -> %s",              \
@@ -455,7 +477,7 @@ static void luv_fs_cb(uv_fs_t* req) {
     }                                                     \
     lua_pushstring(L, uv_err_name(req->result));          \
     if(req->fs_type != UV_FS_SCANDIR) {                   \
-      luv_cleanup_req(L, data);                           \
+      luv_cleanup_req(L, lreq);                           \
       req->data = NULL;                                   \
       uv_fs_req_cleanup(req);                             \
     }                                                     \
@@ -464,13 +486,13 @@ static void luv_fs_cb(uv_fs_t* req) {
   else if (sync) {                                        \
     nargs = push_fs_result(L, req);                       \
     if(req->fs_type != UV_FS_SCANDIR) {                   \
-      luv_cleanup_req(L, data);                           \
+      luv_cleanup_req(L, lreq);                           \
       req->data = NULL;                                   \
       uv_fs_req_cleanup(req);                             \
     }                                                     \
   }                                                       \
   else {                                                  \
-    lua_rawgeti(L, LUA_REGISTRYINDEX, data->req_ref);     \
+    lua_rawgeti(L, LUA_REGISTRYINDEX, lreq->req_ref);     \
     nargs = 1;                                            \
   }                                                       \
 }
@@ -487,7 +509,7 @@ static int luv_fs_close(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(close, req, file);
+  FS_CALL(uv_fs_close, req, file);
 }
 
 static int luv_fs_open(lua_State* L) {
@@ -498,7 +520,7 @@ static int luv_fs_open(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(open, req, path, flags, mode);
+  FS_CALL(uv_fs_open, req, path, flags, mode);
 }
 
 static int luv_fs_read(lua_State* L) {
@@ -508,6 +530,7 @@ static int luv_fs_read(lua_State* L) {
   // -1 offset means "the current file offset is used and updated"
   int64_t offset = -1;
   int ref;
+  char* data;
   // both offset and callback are optional
   if (luv_is_callable(L, 3) && lua_isnoneornil(L, 4)) {
     ref = luv_check_continuation(L, 3);
@@ -516,7 +539,7 @@ static int luv_fs_read(lua_State* L) {
     offset = luaL_optinteger(L, 3, offset);
     ref = luv_check_continuation(L, 4);
   }
-  char* data = (char*)malloc(len);
+  data = (char*)malloc(len);
   if (!data) {
     luaL_unref(L, LUA_REGISTRYINDEX, ref);
     return luaL_error(L, "Failure to allocate buffer");
@@ -526,7 +549,7 @@ static int luv_fs_read(lua_State* L) {
   req->data = luv_setup_req(L, ctx, ref);
   // TODO: find out why we can't just use req->ptr for the base
   ((luv_req_t*)req->data)->data = buf.base;
-  FS_CALL(read, req, file, &buf, 1, offset);
+  FS_CALL(uv_fs_read, req, file, &buf, 1, offset);
 }
 
 static int luv_fs_unlink(lua_State* L) {
@@ -535,7 +558,7 @@ static int luv_fs_unlink(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(unlink, req, path);
+  FS_CALL(uv_fs_unlink, req, path);
 }
 
 static int luv_fs_write(lua_State* L) {
@@ -557,7 +580,7 @@ static int luv_fs_write(lua_State* L) {
   size_t count;
   uv_buf_t* bufs = luv_check_bufs(L, 2, &count, (luv_req_t*)req->data);
   int nargs;
-  FS_CALL_NORETURN(write, req, file, bufs, count, offset);
+  FS_CALL_NORETURN(uv_fs_write, req, file, bufs, count, offset);
   free(bufs);
   return nargs;
 }
@@ -569,7 +592,7 @@ static int luv_fs_mkdir(lua_State* L) {
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(mkdir, req, path, mode);
+  FS_CALL(uv_fs_mkdir, req, path, mode);
 }
 
 static int luv_fs_mkdtemp(lua_State* L) {
@@ -578,7 +601,7 @@ static int luv_fs_mkdtemp(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(mkdtemp, req, tpl);
+  FS_CALL(uv_fs_mkdtemp, req, tpl);
 }
 
 #if LUV_UV_VERSION_GEQ(1, 34, 0)
@@ -588,7 +611,7 @@ static int luv_fs_mkstemp(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(mkstemp, req, tpl);
+  FS_CALL(uv_fs_mkstemp, req, tpl);
 }
 #endif
 
@@ -598,7 +621,7 @@ static int luv_fs_rmdir(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(rmdir, req, path);
+  FS_CALL(uv_fs_rmdir, req, path);
 }
 
 static int luv_fs_scandir(lua_State* L) {
@@ -607,8 +630,34 @@ static int luv_fs_scandir(lua_State* L) {
   int flags = 0; // TODO: find out what these flags are.
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
-  req->data = luv_setup_req_with_mt(L, ctx, ref, "uv_fs");
-  FS_CALL(scandir, req, path, flags);
+  req->data = luv_setup_req(L, ctx, ref);
+  int sync = ref == LUA_NOREF;
+
+  // Wrap the req in a garbage-collectable wrapper.
+  // This allows us to separate the lifetime of the uv_req_t from the lifetime
+  // of the userdata that gets returned here, since otherwise the returned uv_req_t
+  // would hold a reference to itself making it ineligible for garbage collection before
+  // the process ends.
+  luv_fs_scandir_t* scandir_req = (luv_fs_scandir_t*)lua_newuserdata(L, sizeof(luv_fs_scandir_t));
+  scandir_req->req = req;
+  luaL_getmetatable(L, "uv_fs_scandir");
+  lua_setmetatable(L, -2);
+  int scandir_req_index = lua_gettop(L);
+
+  int nargs;
+  FS_CALL_NORETURN(uv_fs_scandir, req, path, flags);
+  // This indicates an error, so we want to return immediately
+  if (nargs != 1) return nargs;
+
+  // Ref the return if this is async, since we don't want this to be garbage collected
+  // before the callback is called.
+  if (!sync) {
+    lua_pushvalue(L, scandir_req_index);
+    ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+  }
+
+  lua_pushvalue(L, scandir_req_index);
+  return 1;
 }
 
 static int luv_fs_scandir_next(lua_State* L) {
@@ -626,7 +675,7 @@ static int luv_fs_stat(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(stat, req, path);
+  FS_CALL(uv_fs_stat, req, path);
 }
 
 static int luv_fs_fstat(lua_State* L) {
@@ -635,7 +684,7 @@ static int luv_fs_fstat(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(fstat, req, file);
+  FS_CALL(uv_fs_fstat, req, file);
 }
 
 static int luv_fs_lstat(lua_State* L) {
@@ -644,7 +693,7 @@ static int luv_fs_lstat(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(lstat, req, path);
+  FS_CALL(uv_fs_lstat, req, path);
 }
 
 static int luv_fs_rename(lua_State* L) {
@@ -657,7 +706,7 @@ static int luv_fs_rename(lua_State* L) {
   // ref the dest path so that we can print it in the error message
   lua_pushvalue(L, 2);
   ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-  FS_CALL(rename, req, path, new_path);
+  FS_CALL(uv_fs_rename, req, path, new_path);
 }
 
 static int luv_fs_fsync(lua_State* L) {
@@ -666,7 +715,7 @@ static int luv_fs_fsync(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(fsync, req, file);
+  FS_CALL(uv_fs_fsync, req, file);
 }
 
 static int luv_fs_fdatasync(lua_State* L) {
@@ -675,7 +724,7 @@ static int luv_fs_fdatasync(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(fdatasync, req, file);
+  FS_CALL(uv_fs_fdatasync, req, file);
 }
 
 static int luv_fs_ftruncate(lua_State* L) {
@@ -685,7 +734,7 @@ static int luv_fs_ftruncate(lua_State* L) {
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(ftruncate, req, file, offset);
+  FS_CALL(uv_fs_ftruncate, req, file, offset);
 }
 
 static int luv_fs_sendfile(lua_State* L) {
@@ -697,7 +746,7 @@ static int luv_fs_sendfile(lua_State* L) {
   int ref = luv_check_continuation(L, 5);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(sendfile, req, out_fd, in_fd, in_offset, length);
+  FS_CALL(uv_fs_sendfile, req, out_fd, in_fd, in_offset, length);
 }
 
 static int luv_fs_access(lua_State* L) {
@@ -707,7 +756,7 @@ static int luv_fs_access(lua_State* L) {
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(access, req, path, amode);
+  FS_CALL(uv_fs_access, req, path, amode);
 }
 
 static int luv_fs_chmod(lua_State* L) {
@@ -717,7 +766,7 @@ static int luv_fs_chmod(lua_State* L) {
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(chmod, req, path, mode);
+  FS_CALL(uv_fs_chmod, req, path, mode);
 }
 
 static int luv_fs_fchmod(lua_State* L) {
@@ -727,7 +776,7 @@ static int luv_fs_fchmod(lua_State* L) {
   int ref = luv_check_continuation(L, 3);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(fchmod, req, file, mode);
+  FS_CALL(uv_fs_fchmod, req, file, mode);
 }
 
 static int luv_fs_utime(lua_State* L) {
@@ -738,7 +787,7 @@ static int luv_fs_utime(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(utime, req, path, atime, mtime);
+  FS_CALL(uv_fs_utime, req, path, atime, mtime);
 }
 
 static int luv_fs_futime(lua_State* L) {
@@ -749,7 +798,7 @@ static int luv_fs_futime(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(futime, req, file, atime, mtime);
+  FS_CALL(uv_fs_futime, req, file, atime, mtime);
 }
 
 #if LUV_UV_VERSION_GEQ(1, 36, 0)
@@ -761,7 +810,7 @@ static int luv_fs_lutime(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(lutime, req, path, atime, mtime);
+  FS_CALL(uv_fs_lutime, req, path, atime, mtime);
 }
 #endif
 
@@ -775,7 +824,7 @@ static int luv_fs_link(lua_State* L) {
   // ref the dest path so that we can print it in the error message
   lua_pushvalue(L, 2);
   ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-  FS_CALL(link, req, path, new_path);
+  FS_CALL(uv_fs_link, req, path, new_path);
 }
 
 static int luv_fs_symlink(lua_State* L) {
@@ -809,7 +858,7 @@ static int luv_fs_symlink(lua_State* L) {
   // ref the dest path so that we can print it in the error message
   lua_pushvalue(L, 2);
   ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-  FS_CALL(symlink, req, path, new_path, flags);
+  FS_CALL(uv_fs_symlink, req, path, new_path, flags);
 }
 
 static int luv_fs_readlink(lua_State* L) {
@@ -818,7 +867,7 @@ static int luv_fs_readlink(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(readlink, req, path);
+  FS_CALL(uv_fs_readlink, req, path);
 }
 
 #if LUV_UV_VERSION_GEQ(1, 8, 0)
@@ -828,7 +877,7 @@ static int luv_fs_realpath(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(realpath, req, path);
+  FS_CALL(uv_fs_realpath, req, path);
 }
 #endif
 
@@ -840,7 +889,7 @@ static int luv_fs_chown(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(chown, req, path, uid, gid);
+  FS_CALL(uv_fs_chown, req, path, uid, gid);
 }
 
 static int luv_fs_fchown(lua_State* L) {
@@ -851,7 +900,7 @@ static int luv_fs_fchown(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(fchown, req, file, uid, gid);
+  FS_CALL(uv_fs_fchown, req, file, uid, gid);
 }
 
 #if LUV_UV_VERSION_GEQ(1, 21, 0)
@@ -863,7 +912,7 @@ static int luv_fs_lchown(lua_State* L) {
   int ref = luv_check_continuation(L, 4);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(lchown, req, path, uid, gid);
+  FS_CALL(uv_fs_lchown, req, path, uid, gid);
 }
 #endif
 
@@ -904,7 +953,7 @@ static int luv_fs_copyfile(lua_State*L) {
   // ref the dest path so that we can print it in the error message
   lua_pushvalue(L, 2);
   ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-  FS_CALL(copyfile, req, path, new_path, flags);
+  FS_CALL(uv_fs_copyfile, req, path, new_path, flags);
 }
 #endif
 
@@ -926,7 +975,7 @@ static int luv_fs_opendir(lua_State* L) {
   lua_pushinteger(L, nentries);
   ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
 
-  FS_CALL(opendir, req, path);
+  FS_CALL(uv_fs_opendir, req, path);
 }
 
 static int luv_fs_readdir(lua_State* L) {
@@ -937,7 +986,12 @@ static int luv_fs_readdir(lua_State* L) {
 
   req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(readdir, req, dir->handle);
+
+  // ref the luv_dir_t so it doesn't get garbage collected before the readdir cb
+  lua_pushvalue(L, 1);
+  ((luv_req_t*)req->data)->data_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+
+  FS_CALL(uv_fs_readdir, req, dir->handle);
 }
 
 static int luv_fs_closedir(lua_State* L) {
@@ -950,7 +1004,7 @@ static int luv_fs_closedir(lua_State* L) {
 
   uv_fs_t *req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(closedir, req, dir->handle);
+  FS_CALL(uv_fs_closedir, req, dir->handle);
 }
 
 static int luv_fs_dir_tostring(lua_State* L) {
@@ -984,7 +1038,7 @@ static int luv_fs_statfs(lua_State* L) {
   int ref = luv_check_continuation(L, 2);
   uv_fs_t* req = (uv_fs_t*)lua_newuserdata(L, uv_req_size(UV_FS));
   req->data = luv_setup_req(L, ctx, ref);
-  FS_CALL(statfs, req, path);
+  FS_CALL(uv_fs_statfs, req, path);
 }
 #endif
 
