@@ -23,9 +23,15 @@ return require('lib/tap')(function (test)
     local constrained = nil
     if uv.get_constrained_memory then
       constrained = uv.get_constrained_memory()
+      assert(constrained >= 0)
+    end
+    local available = nil
+    if uv.get_available_memory then
+      available = uv.get_available_memory()
+      assert(available >= 0, available)
     end
     local free = uv.get_free_memory()
-    p{rss=rss,total=total,free=free, constrained=constrained}
+    p{rss=rss,total=total,free=free,available=available,constrained=constrained}
     assert(rss < total)
   end)
 
@@ -168,4 +174,49 @@ return require('lib/tap')(function (test)
     assert(err:match("^E2BIG"))
   end, "1.33.0")
 
+  test("uv errno", function(print, p, expect, uv)
+    assert(type(uv.errno)=='table')
+    for k, v in pairs(uv.errno) do
+      assert(v < 0, k)
+    end
+  end)
+
+  test("uv constants", function(print, p, expect, uv)
+    assert(type(uv.constants)=='table')
+    for k, v in pairs(uv.constants) do
+      if k=='THREAD_PRIORITY_LOWEST' then
+        assert(v == -2)
+      elseif k=='THREAD_PRIORITY_BELOW_NORMAL' then
+        assert(v == -1)
+      else
+        assert(v >= 0, k)
+      end
+    end
+  end)
+
+  test("uv.cpumask_size", function(print, p, expect, uv)
+    -- The result can vary per-platform and is only supported on some platforms,
+    -- so just test that the function exists and behaves coherently.
+    local size, err = uv.cpumask_size()
+    p(size, err)
+    if err then
+      assert(not size)
+    else
+      assert(size >= 0)
+    end
+  end, "1.45.0")
+
+  test("uv.clock_gettime", function(print, p, expect, uv)
+    for _, clock_id in ipairs({"monotonic", "realtime"}) do
+      local timespec, err = uv.clock_gettime("monotonic")
+      p(clock_id, timespec, err)
+      if err then
+        assert(not timespec)
+      else
+        assert(timespec.sec >= 0)
+        assert(timespec.nsec >= 0)
+      end
+    end
+  end, "1.45.0")
+
 end)
